






fs1D = (12,7)
fs2D = (12,7)

dmargin1D = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                 wspace=0.3, hspace=0.5)
dmargin2D = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                 wspace=0.3, hspace=0.8)
dmargin_combine = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                       wspace=0.3, hspace=0.3)
